﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace VolumeRayCaster
{
    class Renderer
    {
        public Renderer(Form form, Bitmap bitmap, int sizeX, int sizeY)
        {
            this.sizeX = sizeX;
            this.sizeY = sizeY;
            pixels = new Pixel[sizeX+2,sizeY+2];
            this.form = form;
            this.bitmap = bitmap;

            lights = new List<Light>();

            // Domyślne wartości
            fov = DegToRad.Convert(60);
            transformation = new IdentityMatrix();
            world = new EmptyWorld();
            nearClipping = 0.0;
            farClipping = 10.0;
            step = 0.1;
            binaryIters = 15;
            blackCut = 0.03;
            whiteCut = 0.05;
            showProgress = true;
        }
        Vector SimpleRaytrace(Vector startPos, Vector ray, bool enableNearClipping)
        {
            Vector current = startPos;
            if(enableNearClipping)
                current += nearClipping * ray;
            int howMany = (int)((farClipping - nearClipping) / step);
            for (int i = 0; i < howMany; ++i)
            {
                current += ray*step;
                if (world.GetPoint(current))
                    return current;
            }
            return new Vector(double.NaN, double.NaN, double.NaN);
        }
        Vector BinaryRaytrace(Vector prev, Vector current)
        {
            for (int i = 0; i < binaryIters; ++i)
            {
                Vector mid = 0.5 * (prev + current);
                if (world.GetPoint(mid))
                    current = mid;
                else
                    prev = mid;
            }
            return 0.5 * (prev + current);
        }
        bool Collision(Vector startPos, Vector ray, bool enableNearClipping)
        {
            Vector r = SimpleRaytrace(startPos, ray, enableNearClipping);
            return !double.IsNaN(r.x);
        }
        Vector CollisionPoint(Vector startPos, Vector ray, bool enableNearClipping)
        {
            // Szukanie przecięcia promienia ze światem
            Vector p = SimpleRaytrace(startPos, ray, enableNearClipping);

            if (double.IsNaN(p.x))
                return p;

            // Szukanie dokładniej, binarnie "ściany" świata
            Vector prev = p - step*ray;
            return BinaryRaytrace(prev, p);
        }
        public void Render360()
        {
            Matrix baseTransform = transformation;
            for (int i = 0; i < 360; ++i)
            {
                transformation = baseTransform * new YRotationMatrix(Math.PI*i/180);
                Render();
                lock (bitmap)
                {
                    bitmap.Save("render" + i + ".png");
                }
            }
        }
        public void Render360Julia()
        {
            Vector baseVector = new Vector(0.5, 0.4, 0.3);
            Matrix baseTransform = transformation;
            for (int i = 0; i < 360; ++i)
            {
                double rotationAngle = DegToRad.Convert(i);
                double phase = (Math.Cos(rotationAngle) + 1) / 2;
                world = new Julia3D(phase*baseVector,20);
                transformation = baseTransform * new YRotationMatrix(rotationAngle);
                Render();
                lock (bitmap)
                {
                    bitmap.Save("render" + i + ".png");
                }
            }
        }
        public void Render()
        {
            double fovYTan = Math.Tan(fov * 0.5);
            double fovXTan = fovYTan * sizeX / sizeY;

            Vector cameraPos = new Vector() * transformation;

            double minDepth = double.MaxValue;
            double maxDepth = double.MinValue;
            
            // Obliczanie głębokości piksela
            for (int y = 0; y < sizeY+2; ++y)
            {
                for (int x = 0; x < sizeX+2; ++x)
                {
                    // Obliczanie wektora od kamery do danego piksela
                    DirectionVector ray = new DirectionVector();
                    ray.z =  1.0;
                    ray.x =  (2.0 * x / sizeX - 1.0) * fovXTan;
                    ray.y = -(2.0 * y / sizeY - 1.0) * fovYTan;
                    ray *= transformation;
                    ray.Normalize();
                    pixels[x, y] = new Pixel();
                    pixels[x, y].ray = ray;

                    // Śledzenie promienia od kamery w stronę danego piksela
                    Vector collision = CollisionPoint(cameraPos, ray, true);

                    // Zapisanie wyniku - głębokości piksela
                    if (Double.IsNaN(collision.x))
                    {
                        pixels[x, y].depth = double.MaxValue;
                        pixels[x, y].worldPoint = cameraPos + 1000000 * ray;
                    }
                    else
                    {
                        pixels[x, y].depth = (collision - cameraPos).Length;
                        pixels[x, y].worldPoint = collision;

                        minDepth = Math.Min(minDepth, pixels[x, y].depth);
                        maxDepth = Math.Max(maxDepth, pixels[x, y].depth);
                    }

                    // Tymczasowe zapisanie koloru, żeby na ekranie coś się pojawiało
                    pixels[x, y].color = new Vector(1,1,1);
                    pixels[x, y].color *= 1 - (pixels[x, y].depth - nearClipping) / (farClipping - nearClipping);
                }

                if (showProgress)
                    ShowLine(y);
            }

            for (int y = 1; y < sizeY + 1; ++y)
            {
                for (int x = 1; x < sizeX + 1; ++x)
                {
                    if (pixels[x, y].depth > farClipping)
                        continue;

                    // Obliczanie normalnej
                    Vector left = pixels[x - 1, y].worldPoint - pixels[x, y].worldPoint;
                    Vector up = pixels[x, y - 1].worldPoint - pixels[x, y].worldPoint;
                    pixels[x, y].normal = Vector.Cross(left, up);
                    pixels[x, y].normal.Normalize();
                    pixels[x, y].color = pixels[x, y].normal;
                }

                if (showProgress)
                    ShowLine(y);

                for (int x = 1; x < sizeX + 1; ++x)
                {
                    // Resetowanie koloru piksela na czarny
                    pixels[x, y].color = new Vector();
                }
            }
            
            // Obliczanie światła rozproszonego
            foreach (Light l in lights)
            {
                for (int y = 1; y < sizeY + 1; ++y)
                {
                    for (int x = 1; x < sizeX + 1; ++x)
                    {
                        if (pixels[x, y].depth > farClipping)
                            continue;

                        // Sprawdzamy, czy punkt nie jest w cieniu
                        if (Collision(pixels[x, y].worldPoint, -l.direction, false))
                            continue;

                        double intensity = pixels[x, y].normal * -l.direction;
                        if (intensity < 0)
                            intensity = 0;

                        pixels[x, y].color += intensity * l.color;
                    }
                    if (showProgress)
                        ShowLine(y);
                }
            }

            // Obliczanie światła odbitego
            foreach (Light l in lights)
            {
                for (int y = 1; y < sizeY + 1; ++y)
                {
                    for (int x = 1; x < sizeX + 1; ++x)
                    {
                        Pixel pixel = pixels[x, y];
                        if (pixel.depth > farClipping)
                            continue;

                        double dot = -(pixel.normal * pixel.ray);
                        Vector reflected = 2 * dot * pixel.normal + pixel.ray;

                        // Zauważmy, że:
                        // reflected * (-l.direction) == -(reflected * l.direction)
                        double intensity = -(reflected * l.direction);
                        double angle = Math.Acos(intensity) * l.specularSize;
                        if (angle >= Math.PI*0.5)
                            continue;

                        // Sprawdzamy, czy punkt nie jest w cieniu
                        if (Collision(pixels[x, y].worldPoint, reflected, false))
                            continue;

                        intensity = Math.Cos(angle);
                        intensity *= intensity;
                        pixels[x, y].color += intensity * l.color;
                    }
                    if (showProgress)
                        ShowLine(y);
                }
            }

            // Mgła
            double depthAmplitudeInv = 0.8 / (maxDepth - minDepth);
            for (int y = 1; y < sizeY + 1; ++y)
            {
                for (int x = 1; x < sizeX + 1; ++x)
                {
                    double depth = (pixels[x, y].depth - minDepth) * depthAmplitudeInv;
                    pixels[x, y].color *= 1 - depth;
                }
                if (showProgress)
                    ShowLine(y);
            }

            // Prosty blur
            for (int y = 2; y < sizeY + 1; ++y)
            {
                for (int x = 2; x < sizeX + 1; ++x)
                {
                    pixels[x, y].color = 0.5 * pixels[x, y].color + 0.25 * pixels[x - 1, y].color + 0.25 * pixels[x, y - 1].color;
                }
                if (showProgress)
                    ShowLine(y);
            }
            
            // Korekcja kontrastu
            double min = double.MaxValue;
            double max = double.MinValue;
            for (int y = 1; y < sizeY + 1; ++y)
            {
                for (int x = 1; x < sizeX + 1; ++x)
                {
                    Vector color = pixels[x, y].color;
                    double currentMin = color.Min;
                    double currentMax = color.Max;

                    if (currentMax == 0)
                        continue;

                    min = Math.Min(min, currentMin);
                    max = Math.Max(max, currentMax);
                }
            }
            double amplitude = max - min;
            max -= whiteCut * amplitude;
            min += blackCut * amplitude;
            amplitude = max - min;
            double amplitudeInv = 1 / amplitude;
            for (int y = 1; y < sizeY + 1; ++y)
            {
                for (int x = 1; x < sizeX + 1; ++x)
                {
                    Vector color = pixels[x, y].color;
                    color.x = (color.x - min) * amplitudeInv;
                    color.y = (color.y - min) * amplitudeInv;
                    color.z = (color.z - min) * amplitudeInv;
                }
                if (showProgress)
                    ShowLine(y);
            }

            if (!showProgress)
                ShowAll();
        }

        Color ConvertColor(Vector color)
        {
            int r = (int)(255.0 * color.x);
            int g = (int)(255.0 * color.y);
            int b = (int)(255.0 * color.z);

            if (r < 0) r = 0;
            if (r > 255) r = 255;
            if (g < 0) g = 0;
            if (g > 255) g = 255;
            if (b < 0) b = 0;
            if (b > 255) b = 255;

            return Color.FromArgb(r,g,b);
        }
        void ShowAll()
        {
            lock (bitmap)
            {
                for (int y = 0; y < sizeY; ++y)
                {
                    for (int x = 0; x < sizeX; ++x)
                    {
                        bitmap.SetPixel(x, y, ConvertColor(pixels[x, y].color));
                    }
                }
            }
            form.Invalidate();
        }
        void ShowLine(int y)
        {
            if (y == 0 || y == sizeY + 1)
                return;

            lock (bitmap)
            {
                for (int x = 1; x < sizeX + 1; ++x)
                {
                    bitmap.SetPixel(x-1,y-1,ConvertColor(pixels[x,y].color));
                }
            }
            form.Invalidate(new Rectangle(0, y-1, sizeX, 1));
        }

        public List<Light> lights;

        public Pixel[,] pixels;
        Bitmap bitmap;
        int sizeX, sizeY;
        public double fov;
        public Matrix transformation;
        public double nearClipping;
        public double farClipping;
        public double step;
        public int binaryIters;
        public World world;
        public double blackCut;
        public double whiteCut;
        public bool showProgress;

        Form form;
    }
}
